/*
 * This file is kernel module for Inter Galileo Gen 2 to access DRAM as PUF
 *
 * Copyright (C) 2016
 * Authors: Wenjie Xiong <wenjie.xiong@yale.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/workqueue.h>
#include "galmod.h"
#include "io.h"

static unsigned int puf_size = 0;//size of PUF. Where data is written and read from in timer_work_puf
module_param(puf_size, uint, S_IRUGO);
static unsigned int puf_addr = 0;//start address of PUF.
module_param(puf_addr, uint, S_IRUGO);
static unsigned int puf_delaySec = 0;//read the data after "delay seconds"
module_param(puf_delaySec, uint, S_IRUGO);
static unsigned int period = 0;//refresh period in function "timer_work_refresh"
module_param(period, uint, S_IRUGO);
static unsigned int mem=0x2000000;//Mem used by OS. Mem being refreshed in function "timer_work_refresh"
module_param(mem,uint,S_IRUGO);
static unsigned int puf_value=0x0;
module_param(puf_value,uint,S_IRUGO);

static struct delayed_work work_refresh, work_puf;
RegDRCF Drfc_default;
RegDCAL Dcal_default;
RegDPMC0 Dpmc0_default;
RegDECCCTRL Ctr_default;


/*
 This function refresh the part of memory used by OS, every "period" ms.
 It refresh the DRAM by access each row of the memory. Each row of the DRAM is 1024*8
 */
static unsigned int timer_work_refresh(struct work_struct *work)
{
    struct delayed_work *dwork = to_delayed_work(work);
    schedule_delayed_work(dwork, msecs_to_jiffies(period));
    unsigned int addr, data,refresh_stride;
    data = 0;
    refresh_stride =0x800;
    for(addr = 0x10000; addr<0x97000; addr+=refresh_stride){
        data ^= readSysRAM(addr);
    }
    for(addr = 0x98000; addr<0xa0000; addr+=refresh_stride){
        data ^= readSysRAM(addr);
    }
    for(addr = 0x100000; addr<mem; addr+=refresh_stride){
        data ^= readSysRAM(addr);
    }
    return data;
}

/*
 This function read the result of PUF, after stopping the refresh after "delaySec" seconds. And enable the DRAM refresh and ECC for the whole memory.
 */
static unsigned int timer_work_puf(struct work_struct *work)
{
    
    enable_refresh();
    enable_ecc(); 
    if (period > 0) {
        cancel_delayed_work(&work_refresh);
    }
    unsigned int  puf_data;
    printk(KERN_INFO "output_PUF. start addr: 0x%08x, puf_size: 0x%08x delaytime: %d s\n",puf_addr,puf_size, puf_delaySec);
    int i =0;
    for(i = puf_addr;i< puf_addr + puf_size;i+=4){
        puf_data = readRAMbuffer(i);
        printk(KERN_INFO "0x%08x ",puf_data);
    }
    printk(KERN_INFO "End PUF output.\n");
    return 0;
}

/*
 Execute when the module is inserted.
 disable the refresh and ECC of DRAM
 */
int __init galmod_init(void)
{
    printk(KERN_INFO "Module started.\n");
    printk(KERN_INFO "puf_size 0x%x puf addr 0x%x, puf delay %ds.\n Disable self refresh and ecc\n",puf_size,puf_addr,puf_delaySec);
    
    // Store default register values
    Ctr_default.raw = isbR32m(MCU,DECCCTRL);
    Drfc_default.raw = isbR32m(MCU, DRFC);
    Dcal_default.raw = isbR32m(MCU, DCAL);
    Dpmc0_default.raw = isbR32m(MCU, DPMC0);
    printk(KERN_ERR "Ctr: 0x%08x, Drfc: 0x%08x,Dcal: 0x%08x, Dpmc0: 0x%08x\n",Ctr_default.raw,Drfc_default.raw,Dcal_default.raw,Dpmc0_default.raw);
    
    // Check arguments
    if (!puf_size || puf_addr<0x02000000) {
        printk(KERN_ERR "You need to provide all valid parameters\n");
        return -1;
    }
    
    unsigned int  puf_data,puf_old_data;
    puf_data = puf_value;
    printk(KERN_INFO "puf start addr: 0x%08x\n",puf_addr);
    int i =0;
    for(i = puf_addr;i< puf_addr + puf_size;i+=4){
        writeRAMbuffer(i,puf_data);
    }
    printk(KERN_INFO "End.\n");
    
    disable_refresh();
    disable_ecc();
    //start to refresh. schedule "timer_work_refresh"
    if (period > 0) {
        INIT_DELAYED_WORK(&work_refresh, timer_work_refresh);
        schedule_delayed_work(&work_refresh, msecs_to_jiffies(period));
    }
    //schedule "work_puf" to read from puf after delaySec
    if (puf_delaySec) {
        INIT_DELAYED_WORK(&work_puf, timer_work_puf);
        schedule_delayed_work(&work_puf, msecs_to_jiffies(puf_delaySec * 1000));
    }
    printk(KERN_INFO "Exit.msecs_to_jiffies(period)=%d\n",msecs_to_jiffies(period));
    return 0;
}

void __exit galmod_exit(void){
    if(!puf_delaySec && period){
        cancel_delayed_work(&work_refresh);
    }
    printk(KERN_INFO "Remove module.\n");
}

void writeSysRAM(unsigned int write_physaddr, unsigned int data){
    void *write_virtaddr;
    write_virtaddr = phys_to_virt(write_physaddr);
    if(write_virtaddr){
        *((unsigned int*)write_virtaddr) = data;
        iounmap(write_virtaddr);
    }
    else{
        printk(KERN_ERR "phys_to_virt return 0; phyaddr=0x%08x, viraddr=0x%08x\n",write_physaddr, write_virtaddr);
    }
}

unsigned int readSysRAM(unsigned int read_physaddr){
    void *read_virtaddr;
    unsigned int data=-1;
    
    read_virtaddr = phys_to_virt(read_physaddr);
    if (read_virtaddr){
        data = *((unsigned int*)read_virtaddr);
        iounmap(read_virtaddr);
    }
    else{
        printk(KERN_ERR "phys_to_virt return 0; phyaddr=0x%08x, viraddr=0x%08x\n",read_physaddr, read_virtaddr);
    }
    return data;
}

void writeRAMbuffer(unsigned int write_physaddr, unsigned int data){
    void *write_virtaddr;
    write_virtaddr = ioremap(write_physaddr, sizeof(unsigned int));
    if(write_virtaddr){
        *((unsigned int*)write_virtaddr) = data;
        iounmap(write_virtaddr);
    }
    else
        printk(KERN_INFO "ioremap return 0; phyaddr=0x%08x, viraddr=0x%08x\n",write_physaddr, write_virtaddr);
}

unsigned int readRAMbuffer(unsigned int read_physaddr){
    void *read_virtaddr;
    unsigned int data=-1;
    read_virtaddr = ioremap(read_physaddr, sizeof(unsigned int));
    if (read_virtaddr){
        data = *((unsigned int*)read_virtaddr);
        iounmap(read_virtaddr);
    }
    else{
        printk(KERN_ERR "ioremap return 0; phyaddr=0x%08x, viraddr=0x%08x\n",read_physaddr, read_virtaddr);
    }
    return data;
}

void pciwrite32(
                uint32_t bus,
                uint32_t dev,
                uint32_t fn,
                uint32_t reg,
                uint32_t data)
{
    void *write_physaddr;
    void *write_virtaddr;

    write_physaddr = (unsigned int)PCIADDR(bus,dev,fn,reg);   
    unsigned long long write_physaddr64;
    write_physaddr64 = (unsigned int)write_physaddr;   
    write_virtaddr = ioremap_nocache(write_physaddr64, sizeof(unsigned int));

    if (!write_virtaddr) {
        printk(KERN_ERR "Failed to map physical address 0x%08x\n", write_physaddr);
        return -1;
    }
    
   *((unsigned int*)write_virtaddr) = data;
    
    iounmap(write_virtaddr);
    
    return;
}

uint32_t pciread32(
                   uint32_t bus,
                   uint32_t dev,
                   uint32_t fn,
                   uint32_t reg)
{
    void *read_physaddr;
    void *read_virtaddr;
    uint32_t data;
    
    read_physaddr = (unsigned int)PCIADDR(bus,dev,fn,reg);
    unsigned long long read_physaddr64 = (unsigned int)read_physaddr;
    read_virtaddr = ioremap_nocache(read_physaddr64, sizeof(unsigned int));
    if (!read_virtaddr) {
        printk(KERN_ERR "Failed to map physical address 0x%08x\n", read_physaddr);
        return -1;
    }   
    data = *((unsigned int*)read_virtaddr);  
    iounmap(read_virtaddr);

    return data;
}

uint32_t Rd32(
              uint32_t unit,
              uint32_t addr){
    
    pciwrite32(0, 0, 0, SB_PACKET_REG, SB_COMMAND(SB_REG_READ_OPCODE, unit, addr));
    return pciread32(0, 0, 0, SB_DATA_REG);
    
}

void Wr32(
          uint32_t unit,
          uint32_t addr,
          uint32_t data){
    pciwrite32(0, 0, 0, SB_DATA_REG, data);
    pciwrite32(0, 0, 0, SB_PACKET_REG, SB_COMMAND(SB_REG_WRITE_OPCODE, unit, addr));
}

void WrMask32(
              uint32_t unit,
              uint32_t addr,
              uint32_t data,
              uint32_t mask){ return;}

void enable_refresh(void){
    
    isbW32m(MCU, DRFC, Drfc_default.raw);
    isbW32m(MCU, DCAL, Dcal_default.raw);
    isbW32m(MCU, DPMC0, Dpmc0_default.raw);
}

void disable_refresh(void){
    RegDRCF Drfc;
    RegDCAL Dcal;
    RegDPMC0 Dpmc0;
    
    Drfc.raw = isbR32m(MCU, DRFC);
    Drfc.field.tREFI = 0;
    Drfc.field.REFDBTCLR = 0;
    isbW32m(MCU, DRFC, Drfc.raw);
    
    Dcal.raw = isbR32m(MCU, DCAL);
    Dcal.field.ZQCINT = 0;
    isbW32m(MCU, DCAL, Dcal.raw);
    
    Dpmc0.raw = isbR32m(MCU, DPMC0);
    Dpmc0.field.ENPHYCLKGATE = 0;
    Dpmc0.field.DYNSREN = 0;
    isbW32m(MCU, DPMC0, Dpmc0.raw);
}

void enable_ecc(void){
    isbW32m(MCU, DECCCTRL, Ctr_default.raw);
}

void disable_ecc(void){
    RegDECCCTRL Ctr;
    Ctr.raw = isbR32m(MCU,DECCCTRL);
    Ctr.field.SBEEN = 0;
    Ctr.field.DBEEN = 0;
    Ctr.field.ENCBGEN = 0;
    isbW32m(MCU, DECCCTRL, Ctr.raw);
}

module_init(galmod_init);
module_exit(galmod_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wenjie Xiong");
MODULE_DESCRIPTION("Custom module that accesses DRAM PUF");
MODULE_PARM_DESC(puf_size, "size of PUF");
MODULE_PARM_DESC(puf_addr, "The start address to puf unit");
MODULE_PARM_DESC(puf_delaySec, "PUF delay time in seconds");
MODULE_PARM_DESC(period, "The period of the timer in ms. 0 (default) or negative number disables the timer");
MODULE_PARM_DESC(mem, "capacity of memory, in hex, e.g. 32M = 0x2000000");
MODULE_PARM_DESC(puf_value, "initial of PUF");
